Require Export CNF.

Definition Cube := list Literal.

Fixpoint Cubed_CNF (F:CNF) (C:Cube) : CNF :=
  match C with
  | nil => F
  | l :: c => (l :: nil) :: (Cubed_CNF F c)
  end.

Lemma Cubed_CNF_satisfies : forall V F C,
  satisfies V (Cubed_CNF F C) -> satisfies V F.
induction C; simpl; intros; auto.
apply IHC.
apply for_all_satisfies; intros.
eapply satisfies_for_all.
apply H.
auto.
Qed.

Lemma satisfies_Cubed_CNF : forall V (F:CNF) (C:Cube),
  satisfies V F -> (forall c, In c C -> L_satisfies V c) ->
  satisfies V (Cubed_CNF F C).
induction C; simpl; intros; auto.
Qed.

Definition neg_cube (C:Cube) : Clause := map negate C.

Fixpoint noCube (C:list Cube) : CNF :=
  match C with
  | nil => nil
  | (c :: C') => (neg_cube c) :: (noCube C')
  end.

Lemma Literal_neg_neg : forall l, negate (negate l) = l.
induction l; simpl; auto.
Qed.

Lemma CubeAndConquer_lemma : forall Formula Cubes,
  (forall c, In c Cubes -> unsat (Cubed_CNF Formula c)) ->
  unsat (noCube Cubes) -> unsat Formula.
intros.
red; intros; intro.
apply (H0 v); clear H0.
apply for_all_satisfies; intros.
red in H.

assert (In (neg_cube cl) Cubes).
clear H1 v H; induction Cubes; simpl; auto.
inversion_clear H0; auto.
left.
rewrite <- H; clear IHCubes H; induction a; simpl; auto.
rewrite Literal_neg_neg; rewrite <- IHa; auto.

generalize (H _ H2 v); clear H H0 H2; intros.
red in H.

elim (C_satisfies_dec v cl); auto.
intro; elim H; clear H Cubes.
apply for_all_satisfies; intros c' Hcl.
induction cl; simpl in Hcl.
apply satisfies_for_all with Formula; auto.
inversion_clear Hcl.
rewrite <- H; clear H IHcl c'.
simpl; simpl in b.
elim (L_satisfies_dec v (negate a)); auto.
intro; elim b.
left; revert b0; clear b cl H1.
induction a; simpl; elim (v p); auto.
apply IHcl; auto.
intro; apply b.
right; auto.
Qed.
