Require Export Basic.
Require Export Even.
Require Export Bool.
Require Export Recdef.
Require Import Allmaps.
Require Export BinaryTrees.
Require Export BinPos.
Require Export CNF.

Section SetClauses.

Definition SetClause := BinaryTree Literal.

Definition SC_add := BT_add Literal_compare.
Definition SC_in := BT_In (T := Literal).
Definition SC_diff := BT_diff Literal_compare.
Definition SC_wf := BT_wf Literal_compare.

Fixpoint Clause_to_SetClause (cl:Clause) : SetClause :=
  match cl with
  | nil => nought
  | l :: cl' => SC_add l (Clause_to_SetClause cl')
  end.

Lemma C_to_SC_wf : forall (cl:Clause), SC_wf (Clause_to_SetClause cl).
induction cl; simpl; red; simpl; auto.
apply BT_wf_add; auto.
apply Lit_compare_sym_Gt.
apply Lit_compare_sym_Lt.
Qed.

Fixpoint SetClause_to_Clause (cl:SetClause) : Clause :=
  match cl with
  | nought => nil
  | node l cl' cl'' => l :: (SetClause_to_Clause cl') ++ (SetClause_to_Clause cl'')
  end.

Lemma list_append_split : forall A, forall P : A -> Prop,
  forall (l1 l2 w1 w2 : list A),
  (forall x, In x l1 -> P x) /\ (forall x, In x w1 -> P x) /\
  (forall x, In x l2 -> ~P x) /\ (forall x, In x w2 -> ~P x)
  -> (l1 ++ l2) = (w1 ++ w2) -> l1 = w1 /\ l2 = w2.
induction l1; induction w1; simpl; auto; intros.
(* 1/3 *)
exfalso; inversion_clear H; inversion_clear H2; inversion_clear H3.
apply H2 with a; auto.
rewrite H0; left; auto.
(* 2/3 *)
exfalso; inversion_clear H; inversion_clear H2; inversion_clear H3.
apply H4 with a; auto.
rewrite <- H0; left; auto.
(* 3/3 *)
inversion_clear H; inversion_clear H2; inversion_clear H3.
inversion H0.
elim (IHl1 l2 w1 w2); intros; auto.
rewrite H7; rewrite H3; auto.
Qed.

Coercion SetClause_to_Clause : SetClause >-> Clause.

Lemma C_to_SC_In_1 : forall l c, In l c -> SC_in l (Clause_to_SetClause c).
induction c; auto.

simpl; intros.
set (eqL := eq_Lit_compare).
set (Leq := Lit_compare_eq).
inversion_clear H.
rewrite H0; clear H0.
apply BT_add_in; auto.
apply BT_add_mon; auto; apply IHc; auto.
Qed.

Lemma SC_to_C_In_1 : forall l c, In l (SetClause_to_Clause c) -> SC_in l c.
induction c; auto.

simpl; intros.
set (eqL := eq_Lit_compare).
set (Leq := Lit_compare_eq).
inversion_clear H; auto.
elim (in_app_or _ _ _ H0); auto.
Qed.

Lemma C_to_SC_In_2 : forall l c, SC_in l (Clause_to_SetClause c) -> In l c.
induction c; auto.

simpl; intros.
set (eqL := eq_Lit_compare).
set (Leq := Lit_compare_eq).
elim (BT_in_add _ _ _ _ _ H); auto.
Qed.

Lemma SC_to_C_In_2 : forall l c, SC_in l c -> In l (SetClause_to_Clause c).
induction c; auto.

simpl; intros.
set (eqL := eq_Lit_compare).
set (Leq := Lit_compare_eq).
inversion_clear H.
right; apply in_or_app; auto.
inversion_clear H0; auto.
right; apply in_or_app; auto.
Qed.

Lemma SetClause_to_Clause_inv : forall cl cl', SC_wf cl -> SC_wf cl' ->
  SetClause_to_Clause cl = SetClause_to_Clause cl' -> cl = cl'.
induction cl; induction cl'; simpl; intros; auto; inversion H1.
inversion_clear H; inversion_clear H5; inversion_clear H6.
inversion_clear H0; inversion_clear H8; inversion_clear H9.
elim list_append_split with
  (l1 := SetClause_to_Clause cl1) (l2 := SetClause_to_Clause cl2)
  (w1 := SetClause_to_Clause cl'1) (w2 := SetClause_to_Clause cl'2)
  (P := fun x => Literal_compare x t = Lt); intros; auto.
rewrite IHcl1 with cl'1; auto.
rewrite IHcl2 with cl'2; auto.
repeat split; intros.
apply H5; apply SC_to_C_In_1; auto.
rewrite H3; apply H8; apply SC_to_C_In_1; auto.
rewrite H7; [discriminate | apply SC_to_C_In_1; auto].
rewrite H3; rewrite H10; [discriminate | apply SC_to_C_In_1; auto].
Qed.

Lemma SC_satisfies_exists : forall (v:Valuation) (c:SetClause), C_satisfies v c ->
  exists l, SC_in l c /\ L_satisfies v l.
induction c; intros.
inversion H.

inversion_clear H.
exists t; split; simpl; auto.

elim (C_satisfies_exists _ _ H0); intros.
inversion_clear H; exists x; split; auto.
elim (in_app_or _ _ _ H1).
left; apply SC_to_C_In_1; auto.
right; right; apply SC_to_C_In_1; auto.
Qed.

Lemma exists_SC_satisfies : forall (v:Valuation) (c:SetClause) l,
  SC_in l c -> L_satisfies v l -> C_satisfies v c.
induction c; auto.
intros.
apply exists_C_satisfies with l; auto.
apply SC_to_C_In_2; auto.
Qed.

Lemma C_to_C_satisfies_1 : forall (v:Valuation) (c:Clause),
  C_satisfies v c -> C_satisfies v (SetClause_to_Clause (Clause_to_SetClause c)).
intros.
elim (C_satisfies_exists _ _ H); intros.
inversion_clear H0.
apply exists_C_satisfies with x; auto.
apply SC_to_C_In_2; apply C_to_SC_In_1; auto.
Qed.

Lemma C_to_C_satisfies_2 : forall (v:Valuation) (c:Clause),
  C_satisfies v (SetClause_to_Clause (Clause_to_SetClause c)) -> C_satisfies v c.
intros.
elim (C_satisfies_exists _ _ H); intros.
inversion_clear H0.
apply exists_C_satisfies with x; auto.
apply C_to_SC_In_2; apply SC_to_C_In_1; auto.
Qed.

Definition true_SC : SetClause := (node (pos 1) (node (neg 1) nought nought) nought).

Lemma SetClause_eq_dec : forall c c':SetClause, {c = c'} + {c <> c'}.
induction c, c'; auto.
right; discriminate.
right; discriminate.
elim (literal_eq_dec t l); intro Hs;
elim (IHc1 c'1); intro Hc1;
elim (IHc2 c'2); intro Hc2;
try (rewrite Hs; rewrite Hc1; rewrite Hc2; auto);
right; intro; inversion H;
try apply Hc; try apply Hs; auto.
Qed.

Lemma SetClause_eq_nil_cons : forall c: SetClause, {l:Literal & {c':SetClause & {c'':SetClause | c = node l c' c''}}} + {c = nought}.
intro c; case c; auto.
clear c; intros l c; left.
exists l; exists c; exists b; auto.
Qed.

End SetClauses.

Section TreeCNF.

Definition TreeCNF := BinaryTree Clause.

Definition TCNF_in := BT_In (T := Clause).
Definition TCNF_remove := BT_remove Clause_compare : Clause -> TreeCNF -> TreeCNF.
Definition TCNF_wf := BT_wf Clause_compare.
Definition TCNF_join := BT_add_all _ Clause_compare : TreeCNF -> TreeCNF -> TreeCNF.
Definition TCNF_add := BT_add Clause_compare : Clause -> TreeCNF -> TreeCNF.

Lemma TCNF_eq_dec : forall c c':TreeCNF, {c = c'} + {c <> c'}.
induction c, c'; auto; try (right; discriminate).
elim (clause_eq_dec t c); intro H1;
elim (IHc1 c'1); intro H2;
elim (IHc2 c'2); intro H3;
try (rewrite H1; rewrite H2; rewrite H3; auto);
right; intro H; inversion H;
try (apply H1); try (apply H2); auto.
Qed.

Fixpoint T_satisfies (v:Valuation) (c:TreeCNF) : Prop :=
  match c with
  | nought => True
  | node cl c' c'' => (C_satisfies v cl) /\ (T_satisfies v c') /\ (T_satisfies v c'')
  end.

Fixpoint make_TCNF (c:CNF) : TreeCNF :=
  match c with
  | nil => nought
  | cl :: c' => TCNF_add cl (make_TCNF c')
  end.

Lemma make_TCNF_wf : forall c, TCNF_wf (make_TCNF c).
induction c; red; simpl; auto.
apply BT_wf_add; auto.
exact Clause_compare_sym_Gt.
exact Clause_compare_sym_Lt.
Qed.

Lemma satisfies_to_T_satisfies : forall v c,
  satisfies v c -> T_satisfies v (make_TCNF c).
induction c; simpl; auto; intros.
inversion_clear H.
generalize (IHc H1); clear IHc; intro.
set (T := make_TCNF c).
fold T in H; induction T; simpl; intros; auto.
inversion_clear H; inversion_clear H3.
elim Clause_compare; simpl; auto.
Qed.

Lemma T_satisfies_to_satisfies : forall v c,
  T_satisfies v (make_TCNF c) -> satisfies v c.
induction c; simpl; auto.
split.

revert H; clear IHc; set (T := make_TCNF c).
induction T; simpl; auto.
intro; elim H; auto.
set (Z := Clause_compare a t); assert (Z = Clause_compare a t); auto.
induction Z; simpl; intros; inversion_clear H0; inversion_clear H2; auto.
rewrite (eq_Clause_compare _ _ (eq_sym H)); auto.

apply IHc; clear IHc.
revert H; set (T := make_TCNF c).
induction T; simpl; auto.
elim Clause_compare; simpl; intros; inversion_clear H; inversion_clear H1; auto.
Qed.

Fixpoint TCNF_to_CNF (c:TreeCNF) : CNF :=
  match c with
  | nought => nil
  | node cl c1 c2 => TCNF_to_CNF c1 ++ (cl :: TCNF_to_CNF c2)
  end.

Lemma satisfies_to_T_satisfies' : forall v c,
  satisfies v (TCNF_to_CNF c) -> T_satisfies v c.
induction c; simpl; auto; intros.
repeat split; auto.

eapply satisfies_for_all.
apply H.
apply in_or_app; right; simpl; auto.

apply IHc1.
apply for_all_satisfies; intros.
eapply satisfies_for_all.
apply H.
apply in_or_app; auto.

apply IHc2.
apply for_all_satisfies; intros.
eapply satisfies_for_all.
apply H.
apply in_or_app; right; simpl; auto.
Qed.

Lemma T_satisfies_to_satisfies' : forall v c,
  T_satisfies v c -> satisfies v (TCNF_to_CNF c).
induction c; simpl; auto.
intro; inversion_clear H; inversion_clear H1.
apply for_all_satisfies; intros.
elim (in_app_or _ _ _ H1); intros.
eapply satisfies_for_all; [apply IHc1 | idtac]; auto.
inversion_clear H3; auto.
rewrite <- H4; auto.
eapply satisfies_for_all; [apply IHc2 | idtac]; auto.
Qed.

Lemma In_TCNF' : forall cl (c:CNF), In cl c -> TCNF_in cl (make_TCNF c).
generalize eq_Clause_compare.
generalize Clause_compare_eq.
induction c; simpl; auto; intros.
inversion_clear H1.
rewrite H2; apply BT_add_in; auto.
apply BT_add_mon; auto.
apply IHc; auto.
Qed.

Lemma TCNF_In' : forall cl (c:CNF), TCNF_in cl (make_TCNF c) -> In cl c.
induction c; simpl; intros.
inversion H.
elim (BT_in_add _ _ _ _ _ H); intros; auto.
Qed.

Lemma TCNF_In : forall cl c, TCNF_in cl c -> In cl (TCNF_to_CNF c).
induction c; simpl; intros; inversion_clear H.
apply in_or_app; auto.
elim H0; intros; apply in_or_app; simpl; auto.
Qed.

Lemma In_TCNF : forall cl c, In cl (TCNF_to_CNF c) -> TCNF_in cl c.
induction c; simpl; intros; auto.
elim (in_app_or _ _ _ H); intros; auto.
inversion_clear H0; auto.
Qed.

Coercion TCNF_to_CNF : TreeCNF >-> CNF.

End TreeCNF.

Section ICNF.

Definition ICNF := Map {cl:SetClause | SC_wf cl}.

Definition empty_ICNF : ICNF := M0 _.

Fixpoint ICNF_to_TCNF (c:ICNF) : TreeCNF :=
  match c with
  | M0 _ => nought
  | M1 _ _ cl => let (c,_) := cl in node (SetClause_to_Clause c) nought nought
  | M2 _ c' c'' => BT_add_all _ Clause_compare (ICNF_to_TCNF c') (ICNF_to_TCNF c'')
  end.

Coercion ICNF_to_TCNF : ICNF >-> TreeCNF.

Lemma ICNF_to_TCNF_wf : forall (c:ICNF), TCNF_wf (c:TreeCNF).
induction c; red; simpl; auto.
induction a0; simpl; auto.
repeat split; intros; inversion H.
apply BT_wf_add_all; auto.
exact Clause_compare_sym_Gt.
exact Clause_compare_sym_Lt.
Qed.

Lemma ICNF_get_in : forall c (cl:SetClause) i Hc,
  MapGet _ c i = Some (exist SC_wf cl Hc) -> TCNF_in cl (ICNF_to_TCNF c).
induction c; simpl; intros.
inversion H.
revert H; elim BinNat.N.eqb; simpl; intro; inversion H; simpl; auto.

set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).
revert H; induction i; auto.
intro; apply BT_add_all_in; auto; revert H; apply IHc1.
induction p.
intro; apply BT_add_all_mon; auto; revert H; apply IHc2.
intro; apply BT_add_all_in; auto; revert H; apply IHc1.
intro; apply BT_add_all_mon; auto; revert H; apply IHc2.
Qed.

Lemma in_ICNF_get : forall c (cl:SetClause), SC_wf cl ->
  TCNF_in cl (ICNF_to_TCNF c) -> exists i Hc, MapGet _ c i = Some (exist SC_wf cl Hc).
induction c; simpl; intros.
inversion H0.
induction a0; inversion_clear H0; inversion_clear H1.

generalize p; rewrite SetClause_to_Clause_inv with x cl; auto; intros.
exists a; exists p0; elim (BinNat.N.eqb_eq a a); intros.
rewrite H2; simpl; auto.

inversion H0.
elim (BT_in_add_all _ _ _ _ _ H0); intro.
elim (IHc1 _ H H1); intros j Hj.

induction j.
exists BinNums.N0; auto.
exists (BinNums.Npos (BinNums.xO p)); auto.

elim (IHc2 _ H H1); intros j Hj.
induction j.
exists (BinNums.Npos BinNums.xH); auto.
exists (BinNums.Npos (BinNums.xI p)); auto.
Qed.

Lemma in_ICNF_get' : forall c (cl:Clause), TCNF_in cl (ICNF_to_TCNF c) ->
  exists i sc Hsc, MapGet _ c i = Some (exist SC_wf sc Hsc) /\ cl = SetClause_to_Clause sc.
induction c; simpl; intros.
inversion H.
induction a0; inversion_clear H; inversion_clear H0.

exists a; exists x; exists p; elim (BinNat.N.eqb_eq a a); intros.
rewrite H1; simpl; auto.

inversion H.
elim (BT_in_add_all _ _ _ _ _ H); intro.
elim (IHc1 _ H0); intros j Hj.

induction j.
exists BinNums.N0; auto.
exists (BinNums.Npos (BinNums.xO p)); auto.

elim (IHc2 _ H0); intros j Hj.
induction j.
exists (BinNums.Npos BinNums.xH); auto.
exists (BinNums.Npos (BinNums.xI p)); auto.
Qed.

Definition get_ICNF (c:ICNF) i : SetClause :=
  match (MapGet _ c i) with
  | None => true_SC
  | Some cl => let (c,_) := cl in c
  end.

Lemma get_ICNF_in_or_default : forall c i,
  {TCNF_in (get_ICNF c i) (ICNF_to_TCNF c)} + {get_ICNF c i = true_SC}.
induction c; intro i; auto.
induction a0; unfold get_ICNF; simpl; elim BinNat.N.eqb; auto.

set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).

unfold get_ICNF; induction i.
simpl.
elim (IHc1 BinNums.N0); auto.
left; apply BT_add_all_in; auto.

induction p; intros; simpl.
elim (IHc2 (BinNums.Npos p)); auto.
left; apply BT_add_all_mon; auto.

elim (IHc1 (BinNums.Npos p)); auto.
left; apply BT_add_all_in; auto.

elim (IHc2 (BinNums.N0)); auto.
left; apply BT_add_all_mon; auto.
Qed.

Definition del_ICNF i (c:ICNF) : ICNF :=
  MapRemove _ c i.

Lemma ICNF_get_del : forall T c (cl:T) i j,
  MapGet _ (MapRemove _ c j) i = Some cl -> MapGet _ c i = Some cl.
induction c; intros cl i j; auto.

simpl.
set (Haj := BinNat.N.eqb a j); assert (Haj = BinNat.N.eqb a j); auto; clearbody Haj.
set (Hai := BinNat.N.eqb a i); assert (Hai = BinNat.N.eqb a i); auto; clearbody Hai.
revert H H0.
elim Haj; elim Hai; simpl; auto; intros.
inversion H1.
rewrite <- H0 in H1; auto.
rewrite <- H0 in H1; auto.

simpl.
elim BinNat.N.odd; rewrite makeM2_M2;
simpl; induction i; auto.
induction p; auto.
apply IHc2.
apply IHc2.
apply IHc1.
induction p; auto.
apply IHc1.
Qed.

Lemma MapRemove_in : forall c (cl:Clause) j,
  TCNF_in cl (ICNF_to_TCNF (MapRemove _ c j)) -> TCNF_in cl (ICNF_to_TCNF c).
intros.
elim in_ICNF_get' with (MapRemove _ c j) cl; auto.
intros i Hi; inversion_clear Hi; inversion_clear H0; inversion_clear H1.
generalize (ICNF_get_in _ _ _ _ H0); intro.
rewrite H2.
apply ICNF_get_in with i x0.
apply ICNF_get_del with j; auto.
Qed.

Lemma del_ICNF_in : forall (i:ad) (c:ICNF) (cl:Clause),
  TCNF_in cl ((del_ICNF i c):TreeCNF) -> TCNF_in cl (c:TreeCNF).
induction c; auto.
induction a0; intros; simpl.
revert H; simpl.
induction a; elim BinNat.N.eqb; simpl; auto.
intros.
apply MapRemove_in with i; auto.
Qed.

Definition add_ICNF i (cl:SetClause) Hcl (c:ICNF) :=
  MapPut _ c i (exist _ cl Hcl).

Lemma MapPut1_in : forall p i j (ci cj:SetClause) (cl:Clause) Hci Hcj,
  TCNF_in cl (ICNF_to_TCNF (MapPut1 _ i (exist _ ci Hci) j (exist _ cj Hcj) p)) -> cl = ci \/ cl = cj.
induction p; simpl; auto; intros i j ci cj cl;
elim BinNat.N.odd; simpl; auto.
(* 1/6 *)
do 2 intro.
change (TCNF_in cl (BT_add Clause_compare cj (node (SetClause_to_Clause ci) nought nought)) -> cl = ci \/ cl = cj).
intro.
elim (BT_in_add _ _ _ _ _ H); auto.
intro; simpl in H0.
inversion_clear H0; inversion H1; auto.
inversion H0.
(* 2/6 *)
do 2 intro.
change (TCNF_in cl (BT_add Clause_compare ci (node (SetClause_to_Clause cj) nought nought)) -> cl = ci \/ cl = cj).
intro.
elim (BT_in_add _ _ _ _ _ H); auto.
intro; simpl in H0.
inversion_clear H0; inversion H1; auto.
inversion H0.
(* 3/6 *)
intros.
eapply IHp; auto.
apply H.
(* 4/6 *)
intros.
elim (BT_in_add_all _ _ _ _ _ H).
apply IHp; auto.
intro; inversion H0.
(* 5/6 *)
do 2 intro.
change (TCNF_in cl (BT_add Clause_compare cj (node (SetClause_to_Clause ci) nought nought)) -> cl = ci \/ cl = cj).
intro.
elim (BT_in_add _ _ _ _ _ H); auto.
intro; simpl in H0.
inversion_clear H0; inversion H1; auto.
inversion H0.
(* 6/6 *)
do 2 intro.
change (TCNF_in cl (BT_add Clause_compare ci (node (SetClause_to_Clause cj) nought nought)) -> cl = ci \/ cl = cj).
intro.
intros.
elim (BT_in_add _ _ _ _ _ H); auto.
intro; simpl in H0.
inversion_clear H0; inversion H1; auto.
inversion H0.
Qed.

Lemma MapPut_in : forall c (cl:Clause) (cl':SetClause) Hcl' j,
  TCNF_in cl (ICNF_to_TCNF (MapPut _ c j (exist _ cl' Hcl'))) -> cl = cl' \/ TCNF_in cl (ICNF_to_TCNF c).
induction c; simpl; intros.
inversion_clear H; auto.
revert H; elim BinNat.N.lxor; simpl; intros.
inversion_clear H; inversion_clear H0; inversion H; auto.
induction a0.
elim (MapPut1_in _ _ _ _ _ _ _ _ H); auto.
right; right; left; auto.

set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).

revert H; induction j; auto.
simpl; intro.
elim (BT_in_add_all _ _ _ _ _ H); intros.
elim (IHc1 _ _ _ _ H0); intros; auto.
right; apply BT_add_all_in; auto.
right; apply BT_add_all_mon; auto.

case p; simpl; intros; auto.
elim (BT_in_add_all _ _ _ _ _ H); clear H; intros.
right; apply BT_add_all_in; auto.
elim (IHc2 _ _  _ _ H); intros; auto.
right; apply BT_add_all_mon; auto.
elim (BT_in_add_all _ _ _ _ _ H); clear H; intros.
elim (IHc1 _ _  _ _ H); intros; auto.
right; apply BT_add_all_in; auto.
right; apply BT_add_all_mon; auto.
elim (BT_in_add_all _ _ _ _ _ H); clear H; intros.
right; apply BT_add_all_in; auto.
elim (IHc2 _ _ _ _ H); intros; auto.
right; apply BT_add_all_mon; auto.
Qed.

Lemma add_ICNF_unsat : forall i (cl:SetClause) Hcl, forall c:ICNF,
  unsat ((add_ICNF i cl Hcl c):ICNF) -> entails c cl -> unsat c.
intros; intro; intro.
apply H with v.
unfold add_ICNF.
apply for_all_satisfies; intros.
elim (MapPut_in c cl0 cl Hcl i); auto; intros.
rewrite H3; auto.
apply satisfies_for_all with (TCNF_to_CNF (ICNF_to_TCNF c)); auto.
apply TCNF_In; auto.
apply In_TCNF; auto.
Qed.

End ICNF.

Section Propagation.

Lemma SC_in_add : forall (l l':Literal) (cl:SetClause),
  In l (SetClause_to_Clause (SC_add l' cl)) -> l = l' \/ In l (SetClause_to_Clause cl).
induction cl; intros.
inversion H; auto.
elim (BT_in_add _ _ _ _ _ (SC_to_C_In_1 _ _ H)); auto; clear H; intros; inversion_clear H; auto.
elim IHcl1; auto.
right; right; apply in_or_app; auto.
apply SC_to_C_In_2; apply BT_add_mon; auto.
apply eq_Lit_compare.
apply Lit_compare_eq.
inversion_clear H0.
right; left; auto.
elim IHcl2; auto.
right; right; apply in_or_app; auto.
apply SC_to_C_In_2; apply BT_add_mon; auto.
apply eq_Lit_compare.
apply Lit_compare_eq.
Qed.

Lemma propagate_singleton : forall (cs:TreeCNF) (c c':SetClause), forall l,
  SC_diff c c' = (node l nought nought) ->
  entails cs (SetClause_to_Clause (SC_add (negate l) c')) -> entails (TCNF_add c cs) c'.
intros.
red; intros.
set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).

assert (C_satisfies v c).
apply (satisfies_for_all _ _ H1).
apply TCNF_In; apply BT_add_in; auto.

assert (satisfies v cs).
apply for_all_satisfies; intros; apply (satisfies_for_all _ _ H1); auto.
apply TCNF_In; apply BT_add_mon; auto; apply In_TCNF; auto.

elim (C_satisfies_exists _ _ H2); intros l' Hl.
inversion_clear Hl.
elim (BT_diff_in_rev _ _ eq_Lit_compare Lit_compare_eq _ _ _ H l'); intros; auto.
apply exists_C_satisfies with l'; auto.
apply SC_to_C_In_2; auto.
clear H; inversion_clear H6.
inversion H.
inversion_clear H.
generalize (H0 v H3); clear H0.
rewrite <- H6.
intro H0; simpl in H0.
elim (C_satisfies_exists _ _ H0); intros.
inversion_clear H; apply exists_C_satisfies with x; auto.
elim (SC_in_add _ _ _ H7); auto; intros.
elim (L_satisfies_neg v l'); intros.
elim H9; auto; rewrite <- H; auto.
inversion H6.
apply SC_to_C_In_1; auto.
Qed.

Lemma propagate_empty : forall (cs:TreeCNF) (c c':SetClause),
  SC_diff c c' = nought -> entails (TCNF_add c cs) c'.
red; intros.

set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).

assert (C_satisfies v c).
apply (satisfies_for_all _ _ H0).
apply TCNF_In; apply BT_add_in; auto.

assert (satisfies v cs).
apply for_all_satisfies; intros; apply (satisfies_for_all _ _ H0); auto.
apply TCNF_In; apply BT_add_mon; auto; apply In_TCNF; auto.

elim (C_satisfies_exists _ _ H1).
intros l Hl; inversion_clear Hl.
apply exists_C_satisfies with l; auto.
elim (BT_diff_in_rev _ _ eq_Lit_compare Lit_compare_eq _ _ _ H l); intros; auto.
apply SC_to_C_In_2; auto.
inversion H5.
apply SC_to_C_In_1; auto.
Qed.

Lemma propagate_entails : forall (cs:TreeCNF) (c c':SetClause), TCNF_in c cs ->
  entails (TCNF_add c cs) c' -> entails cs c'.
intros.
red; intros.
apply H0; auto.
apply for_all_satisfies; intros.
elim (BT_in_add _ _ _ _ _ (In_TCNF _ _ H2)); intros.
apply satisfies_for_all with (TCNF_to_CNF cs); auto.
apply TCNF_In; auto.
rewrite H3; clear H3 H2 cl.
apply satisfies_for_all with (TCNF_to_CNF cs); auto.
apply TCNF_In; auto.
Qed.

Lemma propagate_true : forall (cs:TreeCNF) (c:SetClause),
  entails (TCNF_add true_SC cs) c -> entails cs c.
intros.
red; intros.
apply H; auto.
apply for_all_satisfies; intros.
elim (BT_in_add _ _ _ _ _ (In_TCNF _ _ H1)); intros.
apply satisfies_for_all with (TCNF_to_CNF cs); auto.
apply TCNF_In; auto.
rewrite H2; apply true_C_true.
Qed.

Fixpoint propagate (cs: ICNF) (c: SetClause) (is:list ad) : bool := 
  match is with
  | nil => false
  | (i::is) => match SetClause_eq_nil_cons (SC_diff (get_ICNF cs i) c) with
    | inright _ => true
    | inleft H => let (l,Hl) := H in let (c',Hc) := Hl in
                  match SetClause_eq_nil_cons c' with
                  | inleft _ => false
                  | inright _ => let (c'',_) := Hc in
                                 match SetClause_eq_nil_cons c'' with
                                 | inleft _ => false
                                 | inright _ => propagate cs (SC_add (negate l) c) is
  end end end end.

Lemma propagate_sound : forall is cs c, propagate cs c is = true -> entails cs c.
induction is; simpl; intros.
inversion H.
revert H.
rename a into i.
elim SetClause_eq_nil_cons; intros.
induction a; induction p; induction p.
revert H; elim SetClause_eq_nil_cons; intros.
inversion H.
rename x into l.
rewrite b in p; clear b x0.
revert H; elim SetClause_eq_nil_cons; intros.
inversion H.
rewrite b in p; clear b x1.
elim (get_ICNF_in_or_default cs i); intro.
apply propagate_entails with (get_ICNF cs i); auto; apply propagate_singleton with l; auto.

apply propagate_true; apply propagate_singleton with l; auto.
rewrite b in p; auto.

elim (get_ICNF_in_or_default cs i); intro.
apply propagate_entails with (get_ICNF cs i); auto; apply propagate_empty; auto.
apply propagate_true; apply propagate_empty; auto.
rewrite b0 in b; auto.
Qed.

End Propagation.

Section Refutation.

Inductive Action : Type :=
  | D : list ad -> Action
  | O : ad -> Clause -> Action
  | R : ad -> Clause -> list ad -> Action.

Definition LazyT := id (A:=Type).

Inductive lazy_list (A:Type) :=
    lnil : lazy_list A
  | lcons : A -> LazyT (lazy_list A) -> lazy_list A.

Definition Oracle := LazyT (lazy_list Action).

Arguments lazy_list [A].
Arguments lcons [A] _ _.

Fixpoint Oracle_size (O:Oracle) : nat :=
  match O with
  | lnil _ => 0
  | lcons (D is) O' => length is + 1 + Oracle_size O'
  | lcons _ O' => 1 + Oracle_size O'
  end.

Definition Answer := bool.

Definition force := id (A := Oracle).
Definition fromVal := id (A := Oracle).

Function refute_work (w:ICNF) (c:TreeCNF) (Hc:TCNF_wf c) (O:Oracle)
  {measure Oracle_size O} : Answer :=
  match (force O) with
  | lnil _ => false
  | lcons (D nil) O' => refute_work w c Hc O'
  | lcons (D (i::is)) O' => refute_work (del_ICNF i w) c Hc (fromVal (lcons (D is) O'))
  | lcons (O i cl) O' => if (BT_in_dec _ _ eq_Clause_compare Clause_compare_eq cl c Hc) then (refute_work (add_ICNF i _ (C_to_SC_wf cl) w) c Hc O') else false
  | lcons (R i nil is) O' => propagate w nought is
  | lcons (R i cl is) O' => andb (propagate w (Clause_to_SetClause cl) is) (refute_work (add_ICNF i _ (C_to_SC_wf cl) w) c Hc O')
  end.
unfold force, id; simpl; intros; rewrite teq; auto with arith.
unfold force, id; simpl; intros; rewrite teq; auto with arith.
unfold force, id; simpl; intros; rewrite teq; auto with arith.
unfold force, id; simpl; intros; rewrite teq; auto with arith.
Defined.

Lemma refute_work_correct : forall w c Hc O, refute_work w c Hc O = true -> unsat (TCNF_join c (w:TreeCNF)).
set (eqC := eq_Clause_compare).
set (Ceq := Clause_compare_eq).
intros w c Hc O; functional induction (refute_work w c Hc O); intros; auto.
(* 1/6 *)
inversion H.
(* 2/6 *)
apply unsat_subset with (TCNF_to_CNF (TCNF_join c ((del_ICNF i w) : TreeCNF))); auto.
intros.
elim (BT_in_add_all Clause _ _ _ _ (In_TCNF _ _ H0)); intro.
apply TCNF_In; apply BT_add_all_in; auto.
apply TCNF_In; apply BT_add_all_mon; auto.
eapply del_ICNF_in; eauto.
(* 3/6 *)
intro; intro; apply IHa with v; auto.
apply for_all_satisfies; intros.
elim (BT_in_add_all _ _ _ _ _ (In_TCNF _ _ H1)); intros; auto.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_in; auto.
elim (MapPut_in _ _ _ _ _ H2); intros.
rewrite H3.
apply C_to_C_satisfies_1.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_in; auto.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_mon; auto.
(* 4/6 *)
inversion H.
(* 5/6 *)
apply unsat_subset with (w:CNF).
intros; apply TCNF_In; apply BT_add_all_mon; auto.
apply In_TCNF; auto.
apply CNF_empty.
replace (nil:Clause) with (SetClause_to_Clause nought); auto.
apply propagate_sound with is; auto.
(* 6/6 *)
elim (andb_true_eq _ _ (eq_sym H)); clear H; intros.
intro; intro; apply IHa with v; auto.
apply for_all_satisfies; intros.
simpl in H2.
elim (BT_in_add_all _ _ _ _ _ (In_TCNF _ _ H2)); intros; auto.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_in; auto.
elim (MapPut_in _ _ _ _ _ H3); intros.
rewrite H4.
apply C_to_C_satisfies_1.
generalize (propagate_sound _ _ _ (eq_sym H)); intro.
apply C_to_C_satisfies_2; apply H5.
apply for_all_satisfies; intros.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_mon; auto.
apply In_TCNF; auto.
apply satisfies_for_all with (TCNF_to_CNF (TCNF_join c (w:TreeCNF))); auto.
apply TCNF_In; apply BT_add_all_mon; auto.
Qed.

Definition refute (c:CNF) (O:Oracle) : Answer :=
  refute_work empty_ICNF (make_TCNF c) (make_TCNF_wf c) O.

Theorem refute_correct : forall c O, refute c O = true -> unsat c.
intros c O; intros; red; intros; intro.
elim (refute_work_correct empty_ICNF (make_TCNF c) (make_TCNF_wf c) O) with v; auto.
apply for_all_satisfies; intros; auto.
apply satisfies_for_all with c; auto.
elim (BT_in_add_all _ _ _ _ _ (In_TCNF _ _ H1)); auto.
apply TCNF_In'.
intro; inversion H2.
Qed.

End Refutation.
